@@ -1,9 +1,9 @@
 /*
  * *****************************************************************************
  *
- * Copyright (c) 2018-2020 Gavin D. Howard and contributors.
+ * SPDX-License-Identifier: BSD-2-Clause
  *
- * All rights reserved.
+ * Copyright (c) 2018-2024 Gavin D. Howard and contributors.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met:
@@ -47,8 +47,6 @@
  * Copyright (c) 2010-2016, Salvatore Sanfilippo <antirez at gmail dot com>
  * Copyright (c) 2010-2013, Pieter Noordhuis <pcnoordhuis at gmail dot com>
  *
- * All rights reserved.
- *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are
  * met:
@@ -81,69 +79,189 @@
 #ifndef BC_HISTORY_H
 #define BC_HISTORY_H
 
-#ifndef BC_ENABLE_HISTORY
-#define BC_ENABLE_HISTORY (1)
-#endif // BC_ENABLE_HISTORY
+// These must come before the #if BC_ENABLE_LINE_LIB below because status.h
+// defines it.
+#include <status.h>
+#include <vector.h>
+
+#if BC_ENABLE_LINE_LIB
+
+#include <stdbool.h>
+#include <setjmp.h>
+#include <signal.h>
+
+extern sigjmp_buf bc_history_jmpbuf;
+extern volatile sig_atomic_t bc_history_inlinelib;
+
+#endif // BC_ENABLE_LINE_LIB
+
+#if BC_ENABLE_EDITLINE
+
+#include <stdio.h>
+#include <histedit.h>
+
+/**
+ * The history struct for editline.
+ */
+typedef struct BcHistory
+{
+	/// A place to store the current line.
+	EditLine* el;
+
+	/// The history.
+	History* hist;
+
+	/// Whether the terminal is bad. This is more or less not used.
+	bool badTerm;
+
+} BcHistory;
+
+// The path to the editrc and its length.
+extern const char bc_history_editrc[];
+extern const size_t bc_history_editrc_len;
+
+#ifdef __APPLE__
+
+/**
+ * Returns true if the line is a valid line, false otherwise.
+ * @param line  The line.
+ * @param len   The length of the line.
+ * @return      True if the line is valid, false otherwise.
+ */
+#define BC_HISTORY_INVALID_LINE(line, len) \
+	((line) == NULL && ((len) == -1 || errno == EINTR))
+
+#else // __APPLE__
+
+/**
+ * Returns true if the line is a valid line, false otherwise.
+ * @param line  The line.
+ * @param len   The length of the line.
+ * @return      True if the line is valid, false otherwise.
+ */
+#define BC_HISTORY_INVALID_LINE(line, len) \
+	((line) == NULL && (len) == -1 && errno == EINTR)
+
+#endif // __APPLE__
+
+#else // BC_ENABLE_EDITLINE
+
+#if BC_ENABLE_READLINE
+
+#include <stdio.h>
+#include <readline/readline.h>
+#include <readline/history.h>
+
+/**
+ * The history struct for readline.
+ */
+typedef struct BcHistory
+{
+	/// A place to store the current line.
+	char* line;
+
+	/// Whether the terminal is bad. This is more or less not used.
+	bool badTerm;
+
+} BcHistory;
+
+#else // BC_ENABLE_READLINE
 
 #if BC_ENABLE_HISTORY
 
-#ifdef _WIN32
-#error History is not supported on Windows.
-#endif
-
-#include <stdbool.h>
 #include <stddef.h>
-#include <stdio.h>
 
 #include <signal.h>
 
+#ifndef _WIN32
 #include <termios.h>
 #include <time.h>
 #include <unistd.h>
 #include <sys/select.h>
+#else // _WIN32
+
+#ifndef WIN32_LEAN_AND_MEAN
+#define WIN32_LEAN_AND_MEAN
+#endif // WIN32_LEAN_AND_MEAN
+
+#include <Windows.h>
+#include <io.h>
+#include <conio.h>
+
+#define strncasecmp _strnicmp
+#define strcasecmp _stricmp
+
+#endif // _WIN32
 
 #include <status.h>
 #include <vector.h>
 #include <read.h>
 
+/// Default columns.
 #define BC_HIST_DEF_COLS (80)
+
+/// Max number of history entries.
 #define BC_HIST_MAX_LEN (128)
+
+/// Max length of a line.
 #define BC_HIST_MAX_LINE (4095)
+
+/// Max size for cursor position buffer.
 #define BC_HIST_SEQ_SIZE (64)
 
+/**
+ * The number of entries in the history.
+ * @param h  The history data.
+ */
 #define BC_HIST_BUF_LEN(h) ((h)->buf.len - 1)
-#define BC_HIST_WRITE(s, n) (write(STDERR_FILENO, (s), (n)) != (ssize_t) (n))
-#define BC_HIST_READ(s, n) (read(STDIN_FILENO, (s), (n)) == -1)
 
+/**
+ * Read n characters into s and check the error.
+ * @param s  The buffer to read into.
+ * @param n  The number of bytes to read.
+ * @return   True if there was an error, false otherwise.
+ */
+#define BC_HIST_READ(s, n) (bc_history_read((s), (n)) == -1)
+
+/// Markers for direction when using arrow keys.
 #define BC_HIST_NEXT (false)
 #define BC_HIST_PREV (true)
 
 #if BC_DEBUG_CODE
-#define lndebug(...)                                               \
-	do {                                                           \
-		if (bc_history_debug_fp == NULL) {                         \
-			bc_history_debug_fp = fopen("/tmp/lndebug.txt","a");   \
-			fprintf(bc_history_debug_fp,                           \
-			       "[%zu %zu %zu] p: %d, rows: %d, "               \
-			       "rpos: %d, max: %zu, oldmax: %d\n",             \
-			       l->len, l->pos, l->oldcolpos, plen, rows, rpos, \
-			       l->maxrows, old_rows);                          \
-		}                                                          \
-		fprintf(bc_history_debug_fp, ", " __VA_ARGS__);            \
-		fflush(bc_history_debug_fp);                               \
-	} while (0)
+
+// These are just for debugging.
+
+#define BC_HISTORY_DEBUG_BUF_SIZE (1024)
+
+// clang-format off
+#define lndebug(...)                                                        \
+	do                                                                      \
+	{                                                                       \
+		if (bc_history_debug_fp.fd == 0)                                    \
+		{                                                                   \
+			bc_history_debug_buf = bc_vm_malloc(BC_HISTORY_DEBUG_BUF_SIZE); \
+			bc_file_init(&bc_history_debug_fp,                              \
+			             open("/tmp/lndebug.txt", O_APPEND),                \
+		                 BC_HISTORY_DEBUG_BUF_SIZE);                        \
+			bc_file_printf(&bc_history_debug_fp,                            \
+			       "[%zu %zu %zu] p: %d, rows: %d, "                        \
+			       "rpos: %d, max: %zu, oldmax: %d\n",                      \
+			       l->len, l->pos, l->oldcolpos, plen, rows, rpos,          \
+			       l->maxrows, old_rows);                                   \
+		}                                                                   \
+		bc_file_printf(&bc_history_debug_fp, ", " __VA_ARGS__);             \
+		bc_file_flush(&bc_history_debug_fp);                                \
+	}                                                                       \
+	while (0)
 #else // BC_DEBUG_CODE
 #define lndebug(fmt, ...)
 #endif // BC_DEBUG_CODE
+// clang-format on
 
-#if !BC_ENABLE_PROMPT
-#define bc_history_line(h, vec, prompt) bc_history_line(h, vec)
-#define bc_history_raw(h, prompt) bc_history_raw(h)
-#define bc_history_edit(h, prompt) bc_history_edit(h)
-#endif // BC_ENABLE_PROMPT
-
-typedef enum BcHistoryAction {
-
+/// An enum of useful actions. To understand what these mean, check terminal
+/// emulators for their shortcuts or the VT100 codes.
+typedef enum BcHistoryAction
+{
 	BC_ACTION_NULL = 0,
 	BC_ACTION_CTRL_A = 1,
 	BC_ACTION_CTRL_B = 2,
@@ -159,12 +277,14 @@ typedef enum BcHistoryAction {
 	BC_ACTION_ENTER = 13,
 	BC_ACTION_CTRL_N = 14,
 	BC_ACTION_CTRL_P = 16,
+	BC_ACTION_CTRL_S = 19,
 	BC_ACTION_CTRL_T = 20,
 	BC_ACTION_CTRL_U = 21,
 	BC_ACTION_CTRL_W = 23,
 	BC_ACTION_CTRL_Z = 26,
 	BC_ACTION_ESC = 27,
-	BC_ACTION_BACKSPACE =  127
+	BC_ACTION_CTRL_BSLASH = 28,
+	BC_ACTION_BACKSPACE = 127
 
 } BcHistoryAction;
 
@@ -172,25 +292,22 @@ typedef enum BcHistoryAction {
  * This represents the state during line editing. We pass this state
  * to functions implementing specific editing functionalities.
  */
-typedef struct BcHistory {
-
+typedef struct BcHistory
+{
 	/// Edited line buffer.
 	BcVec buf;
 
 	/// The history.
 	BcVec history;
 
-	/// A temporary buffer for refresh. Using this
-	/// prevents an allocation on every refresh.
-	BcVec tmp;
+	/// Any material printed without a trailing newline.
+	BcVec extras;
 
-#if BC_ENABLE_PROMPT
 	/// Prompt to display.
-	const char *prompt;
+	const char* prompt;
 
 	/// Prompt length.
 	size_t plen;
-#endif // BC_ENABLE_PROMPT
 
 	/// Prompt column length.
 	size_t pcol;
@@ -207,9 +324,26 @@ typedef struct BcHistory {
 	/// The history index we are currently editing.
 	size_t idx;
 
+#ifndef _WIN32
 	/// The original terminal state.
 	struct termios orig_termios;
+#else // _WIN32
+	///  The original input console mode.
+	DWORD orig_in;
 
+	///  The original output console mode.
+	DWORD orig_out;
+#endif // _WIN32
+
+	/// These next two are here because pahole found a 4 byte hole here.
+
+	/// Whether we are in rawmode.
+	bool rawMode;
+
+	/// Whether the terminal is bad.
+	bool badTerm;
+
+#ifndef _WIN32
 	/// This is to check if stdin has more data.
 	fd_set rdset;
 
@@ -218,37 +352,84 @@ typedef struct BcHistory {
 
 	/// This is to check if stdin has more data.
 	sigset_t sigmask;
-
-	/// This is to signal that there is more, so we don't process yet.
-	bool stdin_has_data;
-
-	/// Whether we are in rawmode.
-	bool rawMode;
-
-	/// Whether the terminal is bad.
-	bool badTerm;
+#endif // _WIN32
 
 } BcHistory;
 
-BcStatus bc_history_line(BcHistory *h, BcVec *vec, const char *prompt);
-void bc_history_add(BcHistory *h, char *line);
+/**
+ * Frees strings used by history.
+ * @param str  The string to free.
+ */
+void
+bc_history_string_free(void* str);
 
-void bc_history_init(BcHistory *h);
-void bc_history_free(BcHistory *h);
+// A list of terminals that don't work.
+extern const char* bc_history_bad_terms[];
 
-extern const char *bc_history_bad_terms[];
+// A tab in history and its length.
 extern const char bc_history_tab[];
 extern const size_t bc_history_tab_len;
+
+// A ctrl+c string.
 extern const char bc_history_ctrlc[];
+
+// UTF-8 data arrays.
 extern const uint32_t bc_history_wchars[][2];
 extern const size_t bc_history_wchars_len;
 extern const uint32_t bc_history_combo_chars[];
 extern const size_t bc_history_combo_chars_len;
+
 #if BC_DEBUG_CODE
-extern FILE *bc_history_debug_fp;
-BcStatus bc_history_printKeyCodes(BcHistory* l);
+
+// Debug data.
+extern BcFile bc_history_debug_fp;
+extern char* bc_history_debug_buf;
+
+/**
+ * A function to print keycodes for debugging.
+ * @param h  The history data.
+ */
+void
+bc_history_printKeyCodes(BcHistory* h);
+
 #endif // BC_DEBUG_CODE
 
 #endif // BC_ENABLE_HISTORY
 
+#endif // BC_ENABLE_READLINE
+
+#endif // BC_ENABLE_EDITLINE
+
+#if BC_ENABLE_HISTORY
+
+/**
+ * Get a line from stdin using history. This returns a status because I don't
+ * want to throw errors while the terminal is in raw mode.
+ * @param h       The history data.
+ * @param vec     A vector to put the line into.
+ * @param prompt  The prompt to display, if desired.
+ * @return        A status indicating an error, if any. Returning a status here
+ *                is better because if we throw an error out of history, we
+ *                leave the terminal in raw mode or in some other half-baked
+ *                state.
+ */
+BcStatus
+bc_history_line(BcHistory* h, BcVec* vec, const char* prompt);
+
+/**
+ * Initialize history data.
+ * @param h  The struct to initialize.
+ */
+void
+bc_history_init(BcHistory* h);
+
+/**
+ * Free history data (and recook the terminal).
+ * @param h  The struct to free.
+ */
+void
+bc_history_free(BcHistory* h);
+
+#endif // BC_ENABLE_HISTORY
+
 #endif // BC_HISTORY_H
